@@ -45,7 +45,7 @@ PMicoDispatcherCallback::callback (CORBA::Dispatcher *dispatcher, CORBA::Dispatc
 
     PUSHMARK(sp);
 
-    XPUSHs(sv_2mortal(newSVpv(ev, 0)));
+    XPUSHs(sv_2mortal(newSVpv((char *)ev, 0)));
     for (int i=0; i<av_len(_args); i++)
 	XPUSHs(sv_2mortal(newSVsv(*av_fetch(_args,i,0))));
 
