@@ -17,6 +17,11 @@ static HV *servant_table = 0;
 
 static const U32 instvars_magic = 0x18981972;
 
+// Used for ExcDescriptionSeq when calling ServantActivator functions
+// that support ForwardRequest
+
+static CORBA::ExcDescriptionSeq *forward_request_seq = NULL;
+
 // Magic (adopted from DBI) to attach InstVars invisibly to perlobj
 PMicoInstVars *
 pmico_instvars_add (SV *perlobj) 
@@ -320,7 +325,7 @@ pmico_oid_to_sv (const PortableServer::ObjectId *oid)
     SvCUR_set(result, oid->length());
     char *pv = SvPVX (result);
     
-    for (int i = 0 ; i < oid->length() ; i++)
+    for (CORBA::ULong i = 0 ; i < oid->length() ; i++)
 	pv[i] = (*oid)[i];
     pv[oid->length()] = '\0';
 
@@ -336,12 +341,30 @@ pmico_sv_to_oid (SV *sv)
     char *str = SvPV(sv, len);
     result = new PortableServer::ObjectId (len);
     result->length(len);
-    for (int i = 0 ; i < len ; i++)
+    for (CORBA::ULong i = 0 ; i < len ; i++)
 	result[i] = str[i];
 
     return result;
 }
 
+static void
+init_forward_request (void)
+{
+    if (!forward_request_seq) {
+
+	forward_request_seq = new CORBA::ExcDescriptionSeq;
+	forward_request_seq->length(1);
+
+	(*forward_request_seq) [0].name = "PortableServer::ForwardRequest";
+	(*forward_request_seq) [0].id =
+	    "IDL:omg.org/PortableServer/ForwardRequest:1.0";
+	(*forward_request_seq) [0].defined_in = "IDL:omg.org/PortableServer:1.0";
+	(*forward_request_seq) [0].version = "1.0";
+	(*forward_request_seq) [0].type = 
+	    new CORBA::TypeCode (*PortableServer::_tc_ForwardRequest);
+    }
+}
+
 CORBA::Boolean
 PMicoAdapterActivator::unknown_adapter (PortableServer::POA_ptr parent, 
 					const char *            name)
@@ -390,6 +413,8 @@ PMicoServantActivator::incarnate (const PortableServer::ObjectId& oid,
     PortableServer::Servant retval;
     CORBA::Exception *exception;
 
+    init_forward_request();
+    
     dSP;
 
     ENTER;
@@ -406,8 +431,7 @@ PMicoServantActivator::incarnate (const PortableServer::ObjectId& oid,
 
     PUTBACK;
 
-    // FIXME: Support ForwardRequest
-    exception = pmico_call_method ("incarnate", 1, NULL);
+    exception = pmico_call_method ("incarnate", 1, forward_request_seq);
 
     if (!exception) {
 	SPAGAIN;
@@ -471,6 +495,8 @@ PMicoServantLocator::preinvoke (const PortableServer::ObjectId& oid,
     PortableServer::Servant retval;
     CORBA::Exception *exception;
 
+    init_forward_request();
+
     dSP;
 
     ENTER;
@@ -489,8 +515,7 @@ PMicoServantLocator::preinvoke (const PortableServer::ObjectId& oid,
 
     PUTBACK;
 
-    // FIXME: Support ForwardRequest
-    exception = pmico_call_method ("preinvoke", 2, NULL);
+    exception = pmico_call_method ("preinvoke", 2, forward_request_seq);
 
     if (!exception) {
 	SPAGAIN;
@@ -563,7 +588,7 @@ PMicoServant::PMicoServant (SV *_perlobj)
     PMicoIfaceInfo *info = pmico_find_interface_description (repoid.c_str());
     
     if (!info)
-	info = pmico_load_interface (NULL, NULL, repoid.c_str());
+	info = pmico_load_contained (NULL, NULL, repoid.c_str());
 
     perlobj = SvRV(_perlobj);
     desc = info->desc;
@@ -632,23 +657,24 @@ PMicoServant::build_args ( const char *name, int &return_items,
     if (!strncmp( name, "_set_", 5)) {
 	CORBA::AttributeDescription *attr_desc = find_attribute(desc, name+5, TRUE);
 	if (attr_desc) {
-	    args = new CORBA::NVList (1);
+	    args = new CORBA::NVList ();
 	    args->add ( CORBA::ARG_IN );
 	    args->item ( 0 )->value()->type( attr_desc->type );
 	}
     } else if (!strncmp( name, "_get_", 5)) {
 	CORBA::AttributeDescription *attr_desc = find_attribute(desc, name+5, FALSE);
 	if (attr_desc) {
-	    args = new CORBA::NVList (0);
+	    args = new CORBA::NVList ();
 	    return_type = attr_desc->type;
 	    return_items++;
 	}
     } else {
 	CORBA::OperationDescription *op_desc = find_operation(desc, name);
 	if (op_desc) {
-	    args = new CORBA::NVList (op_desc->parameters.length());
+	    args = new CORBA::NVList ();
 
 	    for (CORBA::ULong i=0; i<op_desc->parameters.length(); i++) {
+
 		switch (op_desc->parameters[i].mode) {
 		case CORBA::PARAM_IN:
 		    args->add (CORBA::ARG_IN);
@@ -686,8 +712,7 @@ PMicoServant::builtin_invoke (CORBA::ServerRequest_ptr svreq)
 	CORBA::NVList_ptr args = new CORBA::NVList (0);
 	if (svreq->params (args)) {
             CORBA::Any *res = new CORBA::Any;
-            CORBA::Boolean r = (*res <<= pmico_find_interface_description (desc->id)->iface);
-            assert (r);
+            *res <<= pmico_find_interface_description (desc->id)->iface;
             svreq->result (res);
         }
 	return true;
@@ -741,7 +766,6 @@ PMicoServant::invoke ( CORBA::ServerRequest_ptr _req )
 
     for (CORBA::ULong i=0; i<args->count(); i++) {
 	CORBA::Flags dir = args->item(i)->flags();
-	
 	if ((dir == CORBA::ARG_IN) || (dir == CORBA::ARG_INOUT)) {
 	    /* We need the PUTBACK/SPAGAIN here, since the call to
 	     * pmico_from_any might want the stack
